from __future__ import absolute_import, division, print_function, unicode_literals

from ctypes import Structure

from cwinsdk.shared.ntdef import UCHAR

# Command Descriptor Block constants.

CDB6GENERIC_LENGTH = 6
CDB10GENERIC_LENGTH = 10
CDB12GENERIC_LENGTH = 12

SETBITON = 1
SETBITOFF = 0

MODE_PAGE_VENDOR_SPECIFIC = 0x00
MODE_PAGE_ERROR_RECOVERY = 0x01
MODE_PAGE_DISCONNECT = 0x02
MODE_PAGE_FORMAT_DEVICE = 0x03
MODE_PAGE_MRW = 0x03
MODE_PAGE_RIGID_GEOMETRY = 0x04
MODE_PAGE_FLEXIBILE = 0x05
MODE_PAGE_WRITE_PARAMETERS = 0x05
MODE_PAGE_VERIFY_ERROR = 0x07
MODE_PAGE_CACHING = 0x08
MODE_PAGE_PERIPHERAL = 0x09
MODE_PAGE_CONTROL = 0x0A
MODE_PAGE_MEDIUM_TYPES = 0x0B
MODE_PAGE_NOTCH_PARTITION = 0x0C
MODE_PAGE_CD_AUDIO_CONTROL = 0x0E
MODE_PAGE_DATA_COMPRESS = 0x0F
MODE_PAGE_DEVICE_CONFIG = 0x10
MODE_PAGE_XOR_CONTROL = 0x10
MODE_PAGE_MEDIUM_PARTITION = 0x11
MODE_PAGE_ENCLOSURE_SERVICES_MANAGEMENT = 0x14
MODE_PAGE_EXTENDED = 0x15
MODE_PAGE_EXTENDED_DEVICE_SPECIFIC = 0x16
MODE_PAGE_CDVD_FEATURE_SET = 0x18
MODE_PAGE_PROTOCOL_SPECIFIC_LUN = 0x18
MODE_PAGE_PROTOCOL_SPECIFIC_PORT = 0x19
MODE_PAGE_POWER_CONDITION = 0x1A
MODE_PAGE_LUN_MAPPING = 0x1B
MODE_PAGE_FAULT_REPORTING = 0x1C
MODE_PAGE_CDVD_INACTIVITY = 0x1D
MODE_PAGE_ELEMENT_ADDRESS = 0x1D
MODE_PAGE_TRANSPORT_GEOMETRY = 0x1E
MODE_PAGE_DEVICE_CAPABILITIES = 0x1F
MODE_PAGE_CAPABILITIES = 0x2A

MODE_SENSE_RETURN_ALL = 0x3f

MODE_SENSE_CURRENT_VALUES = 0x00
MODE_SENSE_CHANGEABLE_VALUES = 0x40
MODE_SENSE_DEFAULT_VAULES = 0x80
MODE_SENSE_SAVED_VALUES = 0xc0

# 6-byte commands:
SCSIOP_TEST_UNIT_READY = 0x00
SCSIOP_REZERO_UNIT = 0x01
SCSIOP_REWIND = 0x01
SCSIOP_REQUEST_BLOCK_ADDR = 0x02
SCSIOP_REQUEST_SENSE = 0x03
SCSIOP_FORMAT_UNIT = 0x04
SCSIOP_READ_BLOCK_LIMITS = 0x05
SCSIOP_REASSIGN_BLOCKS = 0x07
SCSIOP_INIT_ELEMENT_STATUS = 0x07
SCSIOP_READ6 = 0x08
SCSIOP_RECEIVE = 0x08
SCSIOP_WRITE6 = 0x0A
SCSIOP_PRINT = 0x0A
SCSIOP_SEND = 0x0A
SCSIOP_SEEK6 = 0x0B
SCSIOP_TRACK_SELECT = 0x0B
SCSIOP_SLEW_PRINT = 0x0B
SCSIOP_SET_CAPACITY = 0x0B
SCSIOP_SEEK_BLOCK = 0x0C
SCSIOP_PARTITION = 0x0D
SCSIOP_READ_REVERSE = 0x0F
SCSIOP_WRITE_FILEMARKS = 0x10
SCSIOP_FLUSH_BUFFER = 0x10
SCSIOP_SPACE = 0x11
SCSIOP_INQUIRY = 0x12
SCSIOP_VERIFY6 = 0x13
SCSIOP_RECOVER_BUF_DATA = 0x14
SCSIOP_MODE_SELECT = 0x15
SCSIOP_RESERVE_UNIT = 0x16
SCSIOP_RELEASE_UNIT = 0x17
SCSIOP_COPY = 0x18
SCSIOP_ERASE = 0x19
SCSIOP_MODE_SENSE = 0x1A
SCSIOP_START_STOP_UNIT = 0x1B
SCSIOP_STOP_PRINT = 0x1B
SCSIOP_LOAD_UNLOAD = 0x1B
SCSIOP_RECEIVE_DIAGNOSTIC = 0x1C
SCSIOP_SEND_DIAGNOSTIC = 0x1D
SCSIOP_MEDIUM_REMOVAL = 0x1E

# 10-byte commands
SCSIOP_READ_FORMATTED_CAPACITY = 0x23
SCSIOP_READ_CAPACITY = 0x25
SCSIOP_READ = 0x28
SCSIOP_WRITE = 0x2A
SCSIOP_SEEK = 0x2B
SCSIOP_LOCATE = 0x2B
SCSIOP_POSITION_TO_ELEMENT = 0x2B
SCSIOP_WRITE_VERIFY = 0x2E
SCSIOP_VERIFY = 0x2F
SCSIOP_SEARCH_DATA_HIGH = 0x30
SCSIOP_SEARCH_DATA_EQUAL = 0x31
SCSIOP_SEARCH_DATA_LOW = 0x32
SCSIOP_SET_LIMITS = 0x33
SCSIOP_READ_POSITION = 0x34
SCSIOP_SYNCHRONIZE_CACHE = 0x35
SCSIOP_COMPARE = 0x39
SCSIOP_COPY_COMPARE = 0x3A
SCSIOP_WRITE_DATA_BUFF = 0x3B
SCSIOP_READ_DATA_BUFF = 0x3C
SCSIOP_WRITE_LONG = 0x3F
SCSIOP_CHANGE_DEFINITION = 0x40
SCSIOP_WRITE_SAME = 0x41
SCSIOP_READ_SUB_CHANNEL = 0x42
SCSIOP_UNMAP = 0x42
SCSIOP_READ_TOC = 0x43
SCSIOP_READ_HEADER = 0x44
SCSIOP_REPORT_DENSITY_SUPPORT = 0x44
SCSIOP_PLAY_AUDIO = 0x45
SCSIOP_GET_CONFIGURATION = 0x46
SCSIOP_PLAY_AUDIO_MSF = 0x47
SCSIOP_PLAY_TRACK_INDEX = 0x48
SCSIOP_SANITIZE = 0x48
SCSIOP_PLAY_TRACK_RELATIVE = 0x49
SCSIOP_GET_EVENT_STATUS = 0x4A
SCSIOP_PAUSE_RESUME = 0x4B
SCSIOP_LOG_SELECT = 0x4C
SCSIOP_LOG_SENSE = 0x4D
SCSIOP_STOP_PLAY_SCAN = 0x4E
SCSIOP_XDWRITE = 0x50
SCSIOP_XPWRITE = 0x51
SCSIOP_READ_DISK_INFORMATION = 0x51
SCSIOP_READ_DISC_INFORMATION = 0x51
SCSIOP_READ_TRACK_INFORMATION = 0x52
SCSIOP_XDWRITE_READ = 0x53
SCSIOP_RESERVE_TRACK_RZONE = 0x53
SCSIOP_SEND_OPC_INFORMATION = 0x54
SCSIOP_MODE_SELECT10 = 0x55
SCSIOP_RESERVE_UNIT10 = 0x56
SCSIOP_RESERVE_ELEMENT = 0x56
SCSIOP_RELEASE_UNIT10 = 0x57
SCSIOP_RELEASE_ELEMENT = 0x57
SCSIOP_REPAIR_TRACK = 0x58
SCSIOP_MODE_SENSE10 = 0x5A
SCSIOP_CLOSE_TRACK_SESSION = 0x5B
SCSIOP_READ_BUFFER_CAPACITY = 0x5C
SCSIOP_SEND_CUE_SHEET = 0x5D
SCSIOP_PERSISTENT_RESERVE_IN = 0x5E
SCSIOP_PERSISTENT_RESERVE_OUT = 0x5F

# 12-byte commands
SCSIOP_REPORT_LUNS = 0xA0
SCSIOP_BLANK = 0xA1
SCSIOP_ATA_PASSTHROUGH12 = 0xA1
SCSIOP_SEND_EVENT = 0xA2
SCSIOP_SECURITY_PROTOCOL_IN = 0xA2
SCSIOP_SEND_KEY = 0xA3
SCSIOP_MAINTENANCE_IN = 0xA3
SCSIOP_REPORT_KEY = 0xA4
SCSIOP_MAINTENANCE_OUT = 0xA4
SCSIOP_MOVE_MEDIUM = 0xA5
SCSIOP_LOAD_UNLOAD_SLOT = 0xA6
SCSIOP_EXCHANGE_MEDIUM = 0xA6
SCSIOP_SET_READ_AHEAD = 0xA7
SCSIOP_MOVE_MEDIUM_ATTACHED = 0xA7
SCSIOP_READ12 = 0xA8
SCSIOP_GET_MESSAGE = 0xA8
SCSIOP_SERVICE_ACTION_OUT12 = 0xA9
SCSIOP_WRITE12 = 0xAA
SCSIOP_SEND_MESSAGE = 0xAB
SCSIOP_SERVICE_ACTION_IN12 = 0xAB
SCSIOP_GET_PERFORMANCE = 0xAC
SCSIOP_READ_DVD_STRUCTURE = 0xAD
SCSIOP_WRITE_VERIFY12 = 0xAE
SCSIOP_VERIFY12 = 0xAF
SCSIOP_SEARCH_DATA_HIGH12 = 0xB0
SCSIOP_SEARCH_DATA_EQUAL12 = 0xB1
SCSIOP_SEARCH_DATA_LOW12 = 0xB2
SCSIOP_SET_LIMITS12 = 0xB3
SCSIOP_READ_ELEMENT_STATUS_ATTACHED = 0xB4
SCSIOP_REQUEST_VOL_ELEMENT = 0xB5
SCSIOP_SECURITY_PROTOCOL_OUT = 0xB5
SCSIOP_SEND_VOLUME_TAG = 0xB6
SCSIOP_SET_STREAMING = 0xB6
SCSIOP_READ_DEFECT_DATA = 0xB7
SCSIOP_READ_ELEMENT_STATUS = 0xB8
SCSIOP_READ_CD_MSF = 0xB9
SCSIOP_SCAN_CD = 0xBA
SCSIOP_REDUNDANCY_GROUP_IN = 0xBA
SCSIOP_SET_CD_SPEED = 0xBB
SCSIOP_REDUNDANCY_GROUP_OUT = 0xBB
SCSIOP_PLAY_CD = 0xBC
SCSIOP_SPARE_IN = 0xBC
SCSIOP_MECHANISM_STATUS = 0xBD
SCSIOP_SPARE_OUT = 0xBD
SCSIOP_READ_CD = 0xBE
SCSIOP_VOLUME_SET_IN = 0xBE
SCSIOP_SEND_DVD_STRUCTURE = 0xBF
SCSIOP_VOLUME_SET_OUT = 0xBF
SCSIOP_INIT_ELEMENT_RANGE = 0xE7

# 16-byte commands
SCSIOP_XDWRITE_EXTENDED16 = 0x80
SCSIOP_WRITE_FILEMARKS16 = 0x80
SCSIOP_REBUILD16 = 0x81
SCSIOP_READ_REVERSE16 = 0x81
SCSIOP_REGENERATE16 = 0x82
SCSIOP_EXTENDED_COPY = 0x83
SCSIOP_POPULATE_TOKEN = 0x83
SCSIOP_WRITE_USING_TOKEN = 0x83
SCSIOP_RECEIVE_COPY_RESULTS = 0x84
SCSIOP_RECEIVE_ROD_TOKEN_INFORMATION = 0x84
SCSIOP_ATA_PASSTHROUGH16 = 0x85
SCSIOP_ACCESS_CONTROL_IN = 0x86
SCSIOP_ACCESS_CONTROL_OUT = 0x87
SCSIOP_READ16 = 0x88
SCSIOP_COMPARE_AND_WRITE = 0x89
SCSIOP_WRITE16 = 0x8A
SCSIOP_READ_ATTRIBUTES = 0x8C
SCSIOP_WRITE_ATTRIBUTES = 0x8D
SCSIOP_WRITE_VERIFY16 = 0x8E
SCSIOP_VERIFY16 = 0x8F
SCSIOP_PREFETCH16 = 0x90
SCSIOP_SYNCHRONIZE_CACHE16 = 0x91
SCSIOP_SPACE16 = 0x91
SCSIOP_LOCK_UNLOCK_CACHE16 = 0x92
SCSIOP_LOCATE16 = 0x92
SCSIOP_WRITE_SAME16 = 0x93
SCSIOP_ERASE16 = 0x93
SCSIOP_ZBC_OUT = 0x94
SCSIOP_ZBC_IN = 0x95
SCSIOP_READ_DATA_BUFF16 = 0x9B
SCSIOP_READ_CAPACITY16 = 0x9E
SCSIOP_GET_LBA_STATUS = 0x9E
SCSIOP_GET_PHYSICAL_ELEMENT_STATUS = 0x9E
SCSIOP_REMOVE_ELEMENT_AND_TRUNCATE = 0x9E
SCSIOP_SERVICE_ACTION_IN16 = 0x9E
SCSIOP_SERVICE_ACTION_OUT16 = 0x9F

# 32-byte commands
SCSIOP_OPERATION32 = 0x7F

SERVICE_ACTION_OVERWRITE = 0x01
SERVICE_ACTION_BLOCK_ERASE = 0x02
SERVICE_ACTION_CRYPTO_ERASE = 0x03
SERVICE_ACTION_EXIT_FAILURE = 0x1f

SERVICE_ACTION_XDWRITE = 0x0004
SERVICE_ACTION_XPWRITE = 0x0006
SERVICE_ACTION_XDWRITEREAD = 0x0007
SERVICE_ACTION_WRITE = 0x000B
SERVICE_ACTION_WRITE_VERIFY = 0x000C
SERVICE_ACTION_WRITE_SAME = 0x000D
SERVICE_ACTION_ORWRITE = 0x000E

SERVICE_ACTION_POPULATE_TOKEN = 0x10
SERVICE_ACTION_WRITE_USING_TOKEN = 0x11

SERVICE_ACTION_RECEIVE_TOKEN_INFORMATION = 0x07

SERVICE_ACTION_CLOSE_ZONE = 0x01
SERVICE_ACTION_FINISH_ZONE = 0x02
SERVICE_ACTION_OPEN_ZONE = 0x03
SERVICE_ACTION_RESET_WRITE_POINTER = 0x04

SERVICE_ACTION_REPORT_ZONES = 0x00

REPORT_ZONES_OPTION_LIST_ALL_ZONES = 0x00
REPORT_ZONES_OPTION_LIST_EMPTY_ZONES = 0x01
REPORT_ZONES_OPTION_LIST_IMPLICITLY_OPENED_ZONES = 0x02
REPORT_ZONES_OPTION_LIST_EXPLICITLY_OPENED_ZONES = 0x03
REPORT_ZONES_OPTION_LIST_CLOSED_ZONES = 0x04
REPORT_ZONES_OPTION_LIST_FULL_ZONES = 0x05
REPORT_ZONES_OPTION_LIST_READ_ONLY_ZONES = 0x06
REPORT_ZONES_OPTION_LIST_OFFLINE_ZONES = 0x07

REPORT_ZONES_OPTION_LIST_RWP_ZONES = 0x10
REPORT_ZONES_OPTION_LIST_NON_SEQUENTIAL_WRITE_RESOURCES_ACTIVE_ZONES = 0x11

REPORT_ZONES_OPTION_LIST_NOT_WRITE_POINTER_ZONES = 0x3F

SERVICE_ACTION_READ_CAPACITY16 = 0x10
SERVICE_ACTION_GET_LBA_STATUS = 0x12
SERVICE_ACTION_GET_PHYSICAL_ELEMENT_STATUS = 0x17
SERVICE_ACTION_REMOVE_ELEMENT_AND_TRUNCATE = 0x18

SERVICE_ACTION_REPORT_TIMESTAMP = 0x0F

SERVICE_ACTION_SET_TIMESTAMP = 0x0F

CDB_RETURN_ON_COMPLETION = 0
CDB_RETURN_IMMEDIATE = 1

CDB_FORCE_MEDIA_ACCESS = 0x08

SCSIOP_DENON_EJECT_DISC = 0xE6
SCSIOP_DENON_STOP_AUDIO = 0xE7
SCSIOP_DENON_PLAY_AUDIO = 0xE8
SCSIOP_DENON_READ_TOC = 0xE9
SCSIOP_DENON_READ_SUBCODE = 0xEB

SCSIMESS_ABORT = 0x06
SCSIMESS_ABORT_WITH_TAG = 0x0D
SCSIMESS_BUS_DEVICE_RESET = 0X0C
SCSIMESS_CLEAR_QUEUE = 0X0E
SCSIMESS_COMMAND_COMPLETE = 0X00
SCSIMESS_DISCONNECT = 0X04
SCSIMESS_EXTENDED_MESSAGE = 0X01
SCSIMESS_IDENTIFY = 0X80
SCSIMESS_IDENTIFY_WITH_DISCON = 0XC0
SCSIMESS_IGNORE_WIDE_RESIDUE = 0X23
SCSIMESS_INITIATE_RECOVERY = 0X0F
SCSIMESS_INIT_DETECTED_ERROR = 0X05
SCSIMESS_LINK_CMD_COMP = 0X0A
SCSIMESS_LINK_CMD_COMP_W_FLAG = 0X0B
SCSIMESS_MESS_PARITY_ERROR = 0X09
SCSIMESS_MESSAGE_REJECT = 0X07
SCSIMESS_NO_OPERATION = 0X08
SCSIMESS_HEAD_OF_QUEUE_TAG = 0X21
SCSIMESS_ORDERED_QUEUE_TAG = 0X22
SCSIMESS_SIMPLE_QUEUE_TAG = 0X20
SCSIMESS_RELEASE_RECOVERY = 0X10
SCSIMESS_RESTORE_POINTERS = 0X03
SCSIMESS_SAVE_DATA_POINTER = 0X02
SCSIMESS_TERMINATE_IO_PROCESS = 0X11

SCSIMESS_MODIFY_DATA_POINTER = 0X00
SCSIMESS_SYNCHRONOUS_DATA_REQ = 0X01
SCSIMESS_WIDE_DATA_REQUEST = 0X03

SCSIMESS_MODIFY_DATA_LENGTH = 5
SCSIMESS_SYNCH_DATA_LENGTH = 3
SCSIMESS_WIDE_DATA_LENGTH = 2

class ATA_PASSTHROUGH12(Structure):
	__slots__ = ()
	_fields_ = [
		("OperationCode", UCHAR), # SCSIOP_ATA_PASSTHROUGH12
		("Reserved1", UCHAR, 1),
		("Protocol", UCHAR, 4),
		("MultipleCount", UCHAR, 3),
		("TLength", UCHAR, 2),
		("ByteBlock", UCHAR, 1),
		("TDir", UCHAR, 1),
		("Reserved2", UCHAR, 1),
		("CkCond", UCHAR, 1),
		("Offline", UCHAR, 2),
		("Features", UCHAR),
		("SectorCount", UCHAR),
		("LbaLow", UCHAR),
		("LbaMid", UCHAR),
		("LbaHigh", UCHAR),
		("Device", UCHAR),
		("Command", UCHAR),
		("Reserved3", UCHAR),
		("Control", UCHAR),
	]



from __future__ import absolute_import, division, print_function, unicode_literals

from future.utils import viewkeys
from builtins import sum

from itertools import islice
from typing import TYPE_CHECKING

from .file import PathOrTextIO

if TYPE_CHECKING:
	from typing import Dict, TextIO, Union
	from .gensim import KeyedVectors

from nltk.tokenize import word_tokenize as tokenize

def gensim_indexer(embeddings, doc, ignore=True):
	# type: (KeyedVectors, str) -> Iterator[int]

	for word in tokenize(doc):
		try:
			yield embeddings.vocab[word.lower()].index
		except KeyError:
			if ignore:
				pass
			else:
				raise

def batch_gensim_indexer(embeddings, docs, ignore=True):
	# type: (KeyedVectors, Iterable[str]) -> Iterator[List[int]]

	for doc in docs:
		yield list(gensim_indexer(embeddings, doc, ignore))

def load_freqs(fname, normalize=False, limit=None):
	# type: (Union[str, TextIO], ) -> Dict[str, int]

	with PathOrTextIO(fname, "rt", encoding="utf-8") as fin:
		freqs = dict()

		for line in islice(fin, limit):
			word, count = line.split()
			freqs[word] = int(count)

	if normalize:
		total = sum(viewkeys(freqs))

		for word, count in viewitems(freqs):
			freqs[word] = count / total

	return freqs



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from io import open

from ..exceptions import MalformedFile

class Subtitle(object):

	def __init__(self, start, end, lines):
		self.start = start
		self.end = end
		self.lines = lines

class Sub(object):
	""" MicroDVD subtitle """

	sep = "|"

	def __init__(self, path, mode="r", encoding="utf-8-sig"):
		self.fp = open(path, mode, encoding)
		self.current_line = 0

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		self.close()

	def _readline(self):
		line = next(self.fp).rstrip()
		self.current_line += 1

		start, end, text = line.split("}", 2)
		try:
			start, end = int(start[1:]), int(end[1:])
		except:
			MalformedFile("Error in line {}: sub malformed: {}".format(self.current_line, line))

		return Subtitle(start, end, text.split(self.sep))

	def readline(self):
		try:
			return self._readline()
		except StopIteration:
			return None

	def __iter__(self):
		return self

	def __next__(self):
		return self._readline()

	def write_subtitles(self, subtitle):
		self.fp.write("{{{}}}{{{}}}{}\n".format(subtitle.start, subtitle.end, self.sep.join(subtitle)))

	def close(self):
		self.fp.close()



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import map, range
from math import pow
from random import sample
from typing import TYPE_CHECKING

from .math import reciprocal
from .compat.math import prod

if TYPE_CHECKING:
	from collections.abc import Collection

def sample_range(total, num):
	return sample(range(total), min(num, total))

def arithmetic_mean(col):
	# type: (Collection, ) -> float

	""" Arithmetic mean of `col`. """

	return sum(col) / len(col)

mean = arithmetic_mean

def harmonic_mean(col):
	# type: (Collection, ) -> float

	""" Harmonic mean of `col`. """

	return len(col) / sum(map(reciprocal, col))

def geometric_mean(col):
	# type: (Collection, ) -> float

	""" Geometric mean of `col`. """

	return pow(prod(col), reciprocal(len(col)))



from __future__ import absolute_import, division, print_function, unicode_literals

from bencode import bdecode

from .file import read_file
from .filesystem import FileProperties

def iter_torrent(path):
	""" Returns file informations from a torrent file. 
		Hash (SHA-1) is obtained according to BEP0047 (if available).
	"""

	info = bdecode(read_file(path, "rb"))["info"]

	try:
		files = info["files"]
	except KeyError:
		yield FileProperties(info['name'], info['length'], False, hash=info.get("sha1"))
	else:
		for fd in files:
			path = "/".join(fd['path'])
			yield FileProperties(info['name'] + "/" + path, fd['length'], False, hash=fd.get("sha1"))

if __name__ == "__main__":
	from argparse import ArgumentParser

	parser = ArgumentParser()
	parser.add_argument("torrentfile")
	args = parser.parse_args()

	print(list(iter_torrent(args.torrentfile)))



from __future__ import absolute_import, division, print_function, unicode_literals

class VariableRowMatrix(object):

	def __init__(self):
		self.lol = []

	@classmethod
	def from_list_of_lists(cls, lol):
		# type: (List[List[Number]], ) -> VariableRowMatrix
		m = VariableRowMatrix.__new__(cls)
		m.lol = lol
		return m

	def __len__(self):
		return len(self.lol)

	def __setitem__(self, key, value):
		i, j = key

		try:
			row = self.lol[i]
		except IndexError:
			mul = i - len(self.lol) + 1
			self.lol.extend(mul*[[]])
			row = self.lol[i]

		try:
			row[j] = value
		except IndexError:
			mul = j - len(row) + 1
			row.extend(mul*[0])
			row[j] = value

	def __getitem__(self, key):
		i, j = key
		return self.lol[i][j]

def benchmark(num=9999999, height=1000, width=100):

	import numpy as np
	from genutility.time import PrintStatementTime

	lol = np.arange(height*width).reshape((height, width)).tolist()
	d = dict()
	for i, col in enumerate(lol):
		for j, val in enumerate(col):
			d[i, j] = val

	sm = VariableRowMatrix.from_list_of_lists(lol)

	def index_lol():
		for x, y in zip(xs, ys):
			a = lol[x][y]

	def index_sm():
		for x, y in zip(xs, ys):
			a = sm[x, y]

	def index_sm_2():
		for ind in zip(xs, ys):
			a = sm[ind]

	def index_d():
		for x, y in zip(xs, ys):
			a = d[x, y]

	def index_d_2():
		for ind in zip(xs, ys):
			a = d[ind]

	xs = np.random.randint(0, height, num)
	ys = np.random.randint(0, width, num)

	for i in range(3):
		with PrintStatementTime("list of lists: {delta}"):
			index_lol()
		with PrintStatementTime("VariableRowMatrix: {delta}"):
			index_sm()
		with PrintStatementTime("VariableRowMatrix: {delta}"):
			index_sm_2()
		with PrintStatementTime("dict with tuple key: {delta}"):
			index_d()
		with PrintStatementTime("dict with single key: {delta}"):
			index_d_2()

if __name__ == "__main__":
	benchmark()



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import map, range
from future.utils import viewvalues, raise_from

from sys import version_info
from functools import partial
from math import pi, factorial, sqrt
from collections import defaultdict, Counter
from itertools import islice
from operator import lt, gt, mul
from typing import TYPE_CHECKING

from .compat.math import prod, isqrt, nan as NaN
from .iter import range_count
from .exceptions import EmptyIterable

if TYPE_CHECKING:
	from typing import TypeVar, Comparable, Iterable, Tuple, Optional
	C = TypeVar("C", bound=Comparable)

inf = float("inf")

def reciprocal(n):
	return 1. / n

def sqr(x):
	return x*x

def dot(a, b):
	return sum(map(mul, a, b))

def euclidic_norm(x):
	return sqrt(sum(map(sqr, x)))

def cosine_similarity(a, b):
	return dot(a, b)/euclidic_norm(a)/euclidic_norm(b)

def argmin(l, s=0, e=None):
	if not l:
		raise EmptyIterable()

	arg = s
	e = e or len(l)
	for i in range(s+1, e):
		if l[i] < l[arg]:
			arg = i
	return arg

def argmax(l, s=0, e=None):
	if not l:
		raise EmptyIterable()

	arg = s
	e = e or len(l)
	for i in range(s+1, e):
		if l[i] > l[arg]:
			arg = i
	return arg

def argmax(iterable):
	# type: (Iterable[Orderable], ) -> Tuple[int, Orderable]

	it = iter(iterable)
	arg = 0
	try:
		max = next(it)
	except StopIteration:
		raise_from(EmptyIterable("empty iterable"), None)

	for i, elm in enumerate(it, 1):
		if elm > max:
			arg = i
			max = elm

	return arg

from itertools import count
from operator import itemgetter
def argmax_v2(iterable):
	# type: (Iterable[T], ) -> Tuple[int, T]

	""" nicer, but almost 2 times slower than above"""
	return max(zip(count(), iterable), key=itemgetter(1))[0]

def minmax(a, b):
	# type: (T, T) -> T

	""" `default` argument cannot be used, because the C level function doesn't have a default value
		but an overload.
	"""

	if a <= b:
		return a, b
	else:
		return b, a

def _argfind_cmp(it, target, op1, op2, pos=0):
	# type: (Iterable[C], C, Callable[[C, C], bool], Callable[[C, C], bool], int) -> Tuple[int, C]

	""" Find the largest element in `it` less than or equal to `target` and return the index and value.
		Return (-1, None) if no such element can be found.
	"""

	idx = -1
	val = None

	it = islice(it, pos, None)
	try:
		v = next(it)
	except StopIteration:
		return idx, val

	if v == target:
		return pos, v
	elif op1(v, target):
		idx = pos
		val = v

	for i, v in enumerate(it, pos+1):
		if v == target:
			return i, v
		elif op1(v, target) and (val is None or op2(v, val)):
			idx = i
			val = v

	return idx, val

def argfind_lte(it, target, pos=0):
	return _argfind_cmp(it, target, lt, gt, pos)

def argfind_gte(it, target, pos=0):
	return _argfind_cmp(it, target, gt, lt, pos)

def degree_to_rad(angle):
	# type: (float, ) -> float

	return 2.*pi*angle/360.

def limit(x, left=0, right=1):
	if x >= left:
		if x <= right:
			return x
		else:
			return right
	else:
		return left

def row_indices(n, square_length):
	""" Returns the one-dimensional indices of all the cells belonging to the
		same row as index `n`.
	"""

	j = n // square_length
	return range(j * square_length, (j + 1) * square_length)

def col_indices(n, square_length):
	""" Returns the one-dimensional indices of all the cells belonging to the
		same column as index `n`.
	"""

	j = n % square_length
	return range(j, square_length * (square_length - 1) + j + 1, square_length)

def subblock_indices(n, outer_square_length, inner_square_length):
	# type: (int, int, int) -> Iterator[Tuple[int, int]]

	""" Returns the one-dimensional indices of all the cells belonging to the
		same block as index `n`.
		`outer_square_length` is the number of cells per block.
		`inner_square_length` current has to be sqrt(block_size).
	"""

	i = outer_square_length
	j = inner_square_length

	assert n < i*i, "n not contained in matrix"
	assert j*j == i, "Currently only squares are supported"

	x, y = (n % i // j, n // i // j)

	for iy in range(j):
		for ix in range(j):
			yield (y * i * j) + (i * iy) + (x * j) + (ix)

def relative_luminance(r, g, b):
	# type: (float, float, float) -> float

	""" Converts linear RGB components to relative luminance.
		See: https://en.wikipedia.org/wiki/Relative_luminance
	"""

	return 0.2126*r + 0.7152*g + 0.0722*b

# sys.maxsize
class _PosInfInt():

	def __lt__(self, rhs):
		return False

	def __gt__(self, rhs):
		return True

	def __eq__(self, rhs):
		raise ArithmeticError("don't compare to other PosInfInt...")

	def __add__(self, rhs):
		return self

	def __sub__(self, rhs):
		return self

	def __str__(self):
		return "PosInfInt"

	def __repr__(self):
		return "PosInfInt"

PosInfInt = _PosInfInt()

def multinomial_coefficient(n, ks):
	# type: (int, Iterable[int]) -> int

	return factorial(n) / prod(map(factorial, ks))

def num_unique_permutations(word):
	# type: (str, ) -> int

	""" Number of unique permutations of `word`. """

	return multinomial_coefficient(len(word), viewvalues(Counter(word)))

def discrete_distribution(it):
	# type: (Iterable, ) -> Tuple[defaultdict, int]

	""" Creates a discrete distribution of the elements of `it`. """

	d = defaultdict(int)
	s = 0
	for i in it:
		s += 1
		d[i] += 1
	return (d, s)

def fibonaccigen(f0=0, f1=1):
	# type: (int, int) -> int
	""" A000045 [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, ...]
		arguments are the first two element of the Fibonacci series

		To get a specific Fibonacci number use `fibonacci(n)`.
	"""

	yield f0
	while True:
		yield f1
		f1, f0 = f1+f0, f1

def floordiv2(n):
	while n > 0:
		yield n
		n = n // 2

def _fibonacci(n):
	# returns (F(n), F(n+1)).
	# inspired by https://www.nayuki.io/page/fast-fibonacci-algorithms

	a, b = 0, 1
	for i in reversed(tuple(floordiv2(n))):
		c = a * (b * 2 - a)
		d = a * a + b * b
		if i % 2 == 0:
			a = c
			b = d
		else:
			a = d
			b = c + d
	return a, b

#from .numba import opjit
#@opjit() # doesn't support arbitrary large ints
def _fibonacci_v2(n):
	#see: https://stackoverflow.com/a/1526036
	# 'Structure and Interpretation of Computer Programs' (SICP) ex 1.19
	# http://community.schemewiki.org/?sicp-ex-1.19

	a, b, p, q = 1, 0, 0, 1
	while n > 0:
		if n % 2 == 0: # even
			oldp = p
			p = p*p + q*q
			q = 2*oldp*q + q*q
			n //= 2
		else:
			olda = a
			a = b*q + a*q + a*p
			b = b*p + olda*q
			n -= 1
	return a, b

def fibonacci(n):

	""" implementation of "fast doubling". can be derived from this identity:
		[[1, 1], [1, 0]]^n = [[F_n+1, F_n], [F_n, F_n-1]]
	"""

	if n == 0:
		return 0
	elif n < 0:
		raise ValueError("Negative arguments not implemented")

	return _fibonacci(n-1)[1]

def byte2size(byte, exp=0, base=1024):
	# type: (Number, int, int) -> Tuple[float, str]

	""" Converts integer number to float and unit string """

	byte = float(byte)
	byte_units = ("Byte", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")

	for unit in islice(byte_units, exp, None):
		if (byte >= base):
			byte /= base
		else:
			return byte, unit

def byte2size_str(byte, roundval=3):
	# type: (Number, int) -> str

	""" Converts integer number to human readable size string. """

	byte, value = byte2size(byte)
	if value == "Byte":
		roundval = 0
	return "{:.{}f} {}".format(byte, roundval, value)

def primes(stop=None):
	# type: (Optional[int], ) -> Iterable[int]

	""" Naive algorithm to yield all primes which are smaller than `stop`.
		If `stop` is None it will never stop generating primes.
		For better algorithms see here: https://stackoverflow.com/questions/2211990/how-to-implement-an-efficient-infinite-generator-of-prime-numbers-in-python/10733621#10733621
	"""

	found_primes = []
	first_prime = 2

	for i in range_count(first_prime, stop):
		is_prime = True
		for j in found_primes:
			if j > isqrt(i):
				break
			if i % j == 0:
				is_prime = False
				break

		if is_prime:
			found_primes.append(i)
			yield i

def additaet(n):
	# type: (int, ) -> int
	""" sums all numbers from 0 to n using Gaussian formula.
	basically same performance
		(n*n + n) >> 1
	too slow, floats
		int(((n/2.0) + 0.5) * n)
		int((n + 1) / 2.0 * n)
	too slow, worse complexity
		sum(range(n+1))
	"""

	return ((n + 1) * n) >> 1

addity = additaet

def digitsum(n):
	# type: (int, ) -> int
	""" Use for numbers with more than 6 digits. """

	return sum(map(int, str(n)))

def digitsum_small(n):
	# type: (int, ) -> int
	""" Use for numbers with less than 6 digits. """

	s = 0
	while n:
		s += n % 10
		n //= 10
	return s

def digitsum_base(n, base=10):
	# type: (int, int) -> int
	""" Use with string input if base is not 10. """

	return sum(map(partial(int, base=base), str(n)))

def euclidean_distance(a, b):
	# type: (Number, Number) -> Number

	""" Euclidean distance between `a` and `b`. """

	return abs(b-a)

def closest(numbers, number, distance_func=euclidean_distance):
	# type: (Iterable, Number) -> Number
	""" For a list of `numbers`, return the closest number to `number`. """

	return min(numbers, key=partial(distance_func, number))

if __name__ == "__main__":
	import timeit
	print(min(timeit.repeat(stmt="digitsum(111111)", setup="from __main__ import digitsum")))
	print(min(timeit.repeat(stmt="digitsum_base(111111)", setup="from __main__ import digitsum_base")))
	#print(min(timeit.repeat('_fibonacci_v2(999999999999999999999999999999999)', setup="from __main__ import _fibonacci_v2")))



from __future__ import absolute_import, division, print_function, unicode_literals

import os, fcntl, signal

from .os_shared import _usagetuple

def _lock(fp, exclusive=True, block=False):
	# type: (IO, bool, bool) -> None

	if exclusive:
		operation = fcntl.LOCK_EX
	else:
		operation = fcntl.LOCK_SH

	if not block:
		operation |= fcntl.LOCK_NB

	fcntl.flock(fp, operation)

def _unlock(fp):
	# type: (IO, ) -> None
	fcntl.flock(fp, fcntl.LOCK_UN)

def _disk_usage_posix(path):
	# type: (str, ) -> _usagetuple

	st = os.statvfs(path)

	total = st.f_blocks * st.f_frsize
	free = st.f_bavail * st.f_frsize

	return _usagetuple(total, total-free, free)

def _interrupt_posix():
	os.kill(os.getpid(), signal.SIGINT)

def _filemanager_cmd_posix(path):
	# type: (str, ) -> str

	return "nautilus \"{}\"".format(path) #gnome only. xdg-open for the rest?
